"""This module sets up matplotlib to be used in the other modules."""

import logging

import matplotlib as mpl
import matplotlib.pyplot as plt


# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Can define helper functions here
